/*
Copyright (c) 2008 Hermes <www.elotrolado.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are 
permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of 
  conditions and the following disclaimer. 
- Redistributions in binary form must reproduce the above copyright notice, this list 
  of conditions and the following disclaimer in the documentation and/or other 
  materials provided with the distribution. 
- The names of the contributors may not be used to endorse or promote products derived 
  from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "snd.h"

#ifdef _SDL_LENDIAN_

#else
#include <ogcsys.h>
#endif 

#define VOICE_ACTIVE      128
#define VOICE_DOUBLEBUFFER 64
#define VOICE_SONGSAMPLE    4
#define VOICE_SONGHALT    256   


typedef struct
{
	s32 flags;

	void * voice;
	s32 size;

	void * voice2; // for double buffering
	s32 size2;

	void * voice_aux; // used to restore the voice when you use SND_PlaySongVoice
	s32 size_aux;

	s32 delay; // sample delay
	s32 pitch;
	s32 volume_l;
	s32 volume_r;
	s32 pos;
	s32 sub_pos;

	void (*callback) (s32 voice);

	u32 tick_counter; 

	s32 ticks_for_attack;
	s32 ticks_sub_for_attack;

	s32 ticks_for_sustain;

	s32 ticks_for_decay;
	s32 ticks_sub_for_decay;

	s16 old_sample_l,old_sample_r;

} data_voices;

#ifdef _SDL_LENDIAN_
static s16 audio_buf[2][SND_BUFFERSIZE] __attribute__ ((aligned (16)));
#else
static s16 audio_buf[2][SND_BUFFERSIZE] __attribute__ ((aligned (32)));
#endif
static s32 curr_audio_buf=0;

static s32 audio_init=0;
static s32 audio_paused=1;
static s32 audio_rate=48000;
static s32 audio_rate_div=48;

static u32 sample_counter=0;
static void (*snd_callback)()=NULL;


static data_voices voices[MAX_SND_VOICES];

static s32 * snd_set0w( s32 *p, int n)
{
	while(n>0) {*p++=0;n--;}
	return p;
}

static s8 * snd_set0b( s8 *p, int n)
{
	while(n>0) {*p++=0;n--;}

	return p;
}

#define ADD_SAT_INC(acum, addr, reg) {acum=((s32)(*addr))+(s32)reg;if(acum<-32768) acum=-32768; else if(acum>32767) acum=32767;*addr++=acum;}

#ifdef _SDL_LENDIAN_
#define MIX_METHOD_SET {*buff++=sl;*buff++=sr;}
#define MIX_METHOD_ADD {ADD_SAT_INC(acum, buff, sl);ADD_SAT_INC(acum, buff, sr);}
#else
#define MIX_METHOD_SET {*buff++=sr;*buff++=sl;}
#define MIX_METHOD_ADD {ADD_SAT_INC(acum, buff, sr);ADD_SAT_INC(acum, buff, sl);}
#endif

#define ADJUST_ACUM(pointer, step) \
                     while(voice_acum>=snd_rate) \
	                    {\
	                    voice_acum-=snd_rate;pos+=step;pointer+=step;\
						}



#define MIXER_MACRO(method, pointer, step, getsample1, getsample2) \
  for(n=0;n<size;n++) \
	{ \
	method; \
	\
	voice_acum+=pitch;\
    \
	if(voice_acum>=snd_rate)\
	{\
     \
	 do {\
	    voice_acum-=snd_rate;pos+=step;pointer+=step;\
	    } while(voice_acum>=snd_rate); /* to be sure it is < pitch*/\
	 \
	 if(((void *)pointer)>=end_pointer) \
	   {\
		if(voice->voice2!=NULL && voice->size2>0)\
         {\
         voice->voice=voice->voice2;\
         voice->size=voice->size2;\
         voice->size2=0;\
		 pos=((int) pointer)-((int) end_pointer); /* position in bytes from the start */\
		 end_pointer=((char *) voice->voice)+voice->size;\
		 pos>>= step-1; /* obtain the real index pos*/\
		 pointer=(void *) (((char*)voice->voice));\
		 pointer+=pos;\
	     voice->pos=pos;\
		 }\
		else {size=0;continue;/*break;*/} /* exit if not more samples*/\
	   }\
	 getsample1;getsample2;\
	 \
	}\
	}
    
#define MIXER_MACRO_SONG(method, pointer, step, getsample1, getsample2) \
  for(n=0;n<size;n++) \
	{ \
	method; \
	\
	voice_acum+=pitch;\
    \
	if(voice_acum>=snd_rate)\
	{\
     \
	 do {\
	    voice_acum-=snd_rate;pos+=step;pointer+=step;\
	    } while(voice_acum>=snd_rate); /* to be sure it is < pitch*/\
	 \
	 if(((void *)pointer)>=end_pointer) \
	   {\
		if(voice->voice2!=NULL && voice->size2>0)\
         {\
         voice->voice=voice->voice2;\
         voice->size=voice->size2;\
		 pos=((int) pointer)-((int) end_pointer); /* position in bytes from the start */\
	     end_pointer=((char *) voice->voice)+voice->size;\
		 pos>>= step-1; /* obtain the real index pos*/\
		 pointer=(void *) (((char*)voice->voice));\
		 pointer+=pos;\
	     voice->pos=pos;\
		 }\
		else {sl=sr=0;size=0;voice->flags|=VOICE_SONGHALT;break;} /* exit if not more samples*/\
	   }\
	 if(voice->ticks_for_attack>0) \
	       {\
		   voll=voice->volume_l-(voice->ticks_for_attack>>16);\
		   volr=voice->volume_r-(voice->ticks_for_attack>>16);\
		   if(voll<0) voll=0; if(volr<0) volr=0;\
		   voice->ticks_for_attack-= voice->ticks_sub_for_attack;\
		   }\
	 else {\
	      if(voice->ticks_for_sustain>0) voice->ticks_for_sustain--;\
	      else \
		   {\
		   if(voice->ticks_for_decay>0) \
	        {\
		    voll=voice->volume_l-(voice->volume_l-(voice->ticks_for_decay>>16));\
		    volr=voice->volume_r-(voice->volume_r-(voice->ticks_for_decay>>16));\
		    \
		    if(voll>voice->volume_l) voll=voice->volume_l; if(volr>voice->volume_r) volr=voice->volume_r;\
		    voice->ticks_for_decay-= voice->ticks_sub_for_decay;\
			\
			if(voice->ticks_for_decay<=0) {sl=sr=0;size=0;voice->flags|=VOICE_SONGHALT;continue;}\
		    }\
	       else {sl=sr=0;size=0;voice->flags|=VOICE_SONGHALT;continue;}\
		   }\
		 }\
	 getsample1;getsample2;\
	 \
	}\
	}

static void mix_voices( s16 *buff, s32 samples)
{
register s32 n=0;

s32 size;
register s16 *p2;
register s8  *p1;

register s16 sl=0,sr=0;

register s32 acum;
register s32 voice_acum;
register s32 pitch;
register s32 pos;
register s32 voll;
register s32 volr;
register s32 snd_rate=audio_rate;


void * end_pointer;

s32 samples2=samples;
s16 *buff2=buff;

s32 nvoice;
s32 first=1;

data_voices * voice;



for(nvoice=0;nvoice<MAX_SND_VOICES;nvoice++)
{
voice=&voices[nvoice];
if(!(voice->flags & VOICE_ACTIVE)) continue; 
if(voice->flags & VOICE_SONGHALT) continue; 
	

if(voice->size2==0)
  if(voice->callback) (* voice->callback) (nvoice); 

samples=samples2;
buff=buff2;


/* generate the delay samples */

size=voice->delay;
if(!(voice->flags & VOICE_SONGHALT))
{
if(size>0)
{
if(size>samples) 
	size=samples; 
samples-=size;
voice->delay-=size;

voice->old_sample_l=sl;voice->old_sample_r=sr;
if(first)
	{
	buff= (s16 *) snd_set0w((s32 *) buff,size);
	}
else
 buff+=size<<1;
}
}

/* voice mixer starts here*/

p2=(s16 *) voice->voice;
p1=(s8 *) p2;

pos=voice->pos;




sl=voice->old_sample_l;sr=voice->old_sample_r;

voll=voice->volume_l;
volr=voice->volume_r;

voice_acum=voice->sub_pos;
pitch=voice->pitch;



p1=&p1[pos];
p2=&p2[pos];

while(1)
{
end_pointer=((char *) voice->voice)+voice->size;
switch((voice->flags & 3))
	{
	case VOICE_MONO_8BIT:
		ADJUST_ACUM(p1, 1)
		size=((voice->size)-(pos));
		if(pos>=0 && voice_acum==0) {sl=((s16)p1[0]);sr=sl; sl=(sl)*voll;sr=(sr)*volr;}
		break; 
	
	case VOICE_MONO_16BIT:
		ADJUST_ACUM(p2, 1)
		size=((voice->size)-(pos))>>1;
	    if(pos>=0 && voice_acum==0) {sl=(((s32)p2[0])*voll)>>8;sr=(((s32) p2[0])*volr)>>8;}
		break;
   
	case VOICE_STEREO_8BIT:
		ADJUST_ACUM(p1, 2)
		size=((voice->size>>1)-(pos));
	    if(pos>=0 && voice_acum==0) {sl=((s16)p1[0])*voll;sr=((s16) p1[1])*volr;}
		break;
   
    case VOICE_STEREO_16BIT:
		ADJUST_ACUM(p2, 2)
		size=((voice->size>>1)-(pos))>>1;
	    if(pos>=0 && voice_acum==0){sl=(((s32)p2[0])*voll)>>8;sr=(((s32) p2[1])*volr)>>8;}
		
		break;
	}



if(size>0) break;

if(voice->voice2!=NULL && voice->size2>0)
         {
         voice->voice=voice->voice2;
         voice->size=voice->size2;
         if(!(voice->flags & VOICE_SONGSAMPLE)) voice->size2=0;
		 voice->pos=pos=0,
		 p2=(s16 *) voice->voice;
		 p1=(s8 *) p2;
		 if(voice->size2==0)
            if(voice->callback) (* voice->callback) (nvoice); 
		 }
else break;
}

size=samples; 

// voice mixer loop
n=0;

switch((voice->flags & 7))
	{
	case VOICE_MONO_8BIT:
		if(first)
		  {
		  if(voll==255 && volr==255) // minor cpu use
			 {MIXER_MACRO(MIX_METHOD_SET, p1, 1, sr=sl=((s16)p1[0])<<8, /* space */ )}
		  else {MIXER_MACRO(MIX_METHOD_SET, p1, 1, sr=sl=((s16)p1[0]) , sl*=voll;sr*=volr)}
		  }
		 else
		  {
		  if(voll==255 && volr==255) // minor cpu use
			 {MIXER_MACRO(MIX_METHOD_ADD, p1, 1, sr=sl=((s16)p1[0])<<8,  /* space */ )}
		  else {MIXER_MACRO(MIX_METHOD_ADD, p1, 1, sr=sl=((s16)p1[0]), sl*=voll;sr*=volr)}
		  }
		break; 
	
	case VOICE_MONO_16BIT:
		if(first)
		  {
		  if(voll==255 && volr==255) // minor cpu use
			 {MIXER_MACRO(MIX_METHOD_SET, p2, 1, sl=sr=p2[0],  /* space */ )}
		  else {MIXER_MACRO(MIX_METHOD_SET, p2, 1, sl=sr=p2[0], sl=(((s32) sl)*voll)>>8;sr=(((s32) sr)*volr)>>8)}
          }
		 else
		  {
		  if(voll==255 && volr==255) // minor cpu use
			 {MIXER_MACRO(MIX_METHOD_ADD, p2, 1, sl=sr=p2[0],  /* space */ )}
		  else {MIXER_MACRO(MIX_METHOD_ADD, p2, 1, sl=sr=p2[0], sl=(((s32) sl)*voll)>>8;sr=(((s32) sr)*volr)>>8)}
          }
		break;
   
	case VOICE_STEREO_8BIT:
		if(first)
		  {
		  if(voll==255 && volr==255) // minor cpu use
			 {MIXER_MACRO(MIX_METHOD_SET, p1, 2, sl=((s16)p1[0]), sr=((s16) p1[1]))}
		  else {MIXER_MACRO(MIX_METHOD_SET, p1, 2, sl=((s16)p1[0])*voll, sr=((s16) p1[1])*volr)}
		  }
		 else
		  {
		  if(voll==255 && volr==255) // minor cpu use
			 {MIXER_MACRO(MIX_METHOD_ADD, p1, 2, sl=((s16)p1[0]), sr=((s16) p1[1]))}
		  else {MIXER_MACRO(MIX_METHOD_ADD, p1, 2, sl=((s16)p1[0])*voll, sr=((s16) p1[1])*volr)}
		  }
		break;
	
    case VOICE_STEREO_16BIT:
		if(first)
		  {
		  if(voll==255 && volr==255)
			 { 
		     MIXER_MACRO(MIX_METHOD_SET, p2, 2, sl=p2[0], sr=p2[1]) // minor cpu use
			 }
		  else
			 {MIXER_MACRO(MIX_METHOD_SET, p2, 2, sl=(((s32)p2[0])*voll)>>8, sr=(((s32) p2[1])*volr)>>8)}
		  }
		 else
		  {
		  if(voll==255 && volr==255)
			 {MIXER_MACRO(MIX_METHOD_ADD, p2, 2, sl=p2[0], sr=p2[1])}
		  else
			 {MIXER_MACRO(MIX_METHOD_ADD, p2, 2, sl=(((s32)p2[0])*voll)>>8, sr=(((s32) p2[1])*volr)>>8)}
		  }
	
		break;

	case VOICE_SONGSAMPLE + VOICE_MONO_8BIT:
	
		if(!(voice->flags & VOICE_SONGHALT))
		{
		if(first)
		  {
		   {MIXER_MACRO_SONG(MIX_METHOD_SET, p1, 1, sr=sl=((s16)p1[0]) , sl*=voll;sr*=volr)}
		  }
		 else
		  {
		   {MIXER_MACRO_SONG(MIX_METHOD_ADD, p1, 1, sr=sl=((s16)p1[0]), sl*=voll;sr*=volr)}
		  }
		}
		
		break; 

	case VOICE_SONGSAMPLE + VOICE_MONO_16BIT:
		if(!(voice->flags & VOICE_SONGHALT))
		{
		if(first)
		  {
		   {MIXER_MACRO_SONG(MIX_METHOD_SET, p2, 1, sl=sr=p2[0], sl=(((s32) sl)*voll)>>8;sr=(((s32) sr)*volr)>>8)}
          }
		 else
		  {
		   {MIXER_MACRO_SONG(MIX_METHOD_ADD, p2, 1, sl=sr=p2[0], sl=(((s32) sl)*voll)>>8;sr=(((s32) sr)*volr)>>8)}
          }
		}
		break;

	}

voice->tick_counter+=(u32) n;
samples-=n; // n is the number of samples  sended from MIXER_MACRO


if(voice->size2==0)
    if(voice->callback) (* voice->callback) (nvoice); 



if(samples) // remain samples?
{if(!(voice->flags & VOICE_DOUBLEBUFFER)) voice->flags&=~VOICE_ACTIVE; // disable the voice 
if(first) // set to 0 the rest if it is the first voice
	{
	buff= (s16 *) snd_set0w((s32 *) buff,samples);
	}
sl=sr=0;
}
voice->old_sample_l=sl;voice->old_sample_r=sr;
voice->pos=pos;
voice->sub_pos=voice_acum;
first=0;
} // bucle de voces

if(first)
	{
    snd_set0w((s32 *) buff2,samples2);
	}
}

#ifdef _SDL_LENDIAN_
static void audio_dma_callback(void *unused, Uint8 *stream, int len)
{	
memcpy((void *) stream, (void *) audio_buf[curr_audio_buf],SND_BUFFERSIZE<<1);

curr_audio_buf ^= 1;
audio_paused=0;

if(snd_callback)
	 (*snd_callback)();

mix_voices(audio_buf[curr_audio_buf], SND_BUFFERSIZE>>2);

sample_counter+=SND_BUFFERSIZE>>2;

if(snd_callback)
	 (*snd_callback)();

mix_voices(((s16 *)audio_buf[curr_audio_buf])+(SND_BUFFERSIZE>>1), SND_BUFFERSIZE>>2);

sample_counter+=SND_BUFFERSIZE>>2;
}
#else

static void audio_dma_callback()
{	

AUDIO_StopDMA();
AUDIO_InitDMA((u32)audio_buf[curr_audio_buf],SND_BUFFERSIZE<<1);
AUDIO_StartDMA();

curr_audio_buf ^= 1;
audio_paused=0;
/*
if(snd_callback)
	 (*snd_callback)();

mix_voices(audio_buf[curr_audio_buf], SND_BUFFERSIZE>>1);
DCFlushRange(audio_buf[curr_audio_buf],SND_BUFFERSIZE<<1);
sample_counter+=SND_BUFFERSIZE>>1;
*/
if(snd_callback)
	 (*snd_callback)();

mix_voices(audio_buf[curr_audio_buf], SND_BUFFERSIZE>>2);

sample_counter+=SND_BUFFERSIZE>>2;

if(snd_callback)
	 (*snd_callback)();

mix_voices(((s16 *)audio_buf[curr_audio_buf])+(SND_BUFFERSIZE>>1), SND_BUFFERSIZE>>2);
DCFlushRange(audio_buf[curr_audio_buf],SND_BUFFERSIZE<<1);
sample_counter+=SND_BUFFERSIZE>>2;

}
#endif

// SOUND FUNCTIONS

#ifdef _SDL_LENDIAN_
SDL_AudioSpec audio;
#endif

int SND_GetAudioRate()
{
return audio_rate;
}

void SND_Init(s32 samplerate)
{
#ifdef _SDL_LENDIAN_
audio_init=1;
#else
if(audio_init==0)
{
AUDIO_Init(NULL);
audio_init=1;
}

AUDIO_StopDMA();
#endif

sample_counter=0;
snd_callback=NULL;

snd_set0b((s8 *) voices,sizeof(data_voices) * MAX_SND_VOICES);

audio_paused=1;
#ifdef _SDL_LENDIAN_
if(samplerate==0) {audio_rate=44100;}
else {audio_rate=48000;}

audio.freq=audio_rate;
audio.format=AUDIO_S16;
audio.channels=2;

audio.samples=SND_BUFFERSIZE/2;
audio.callback = audio_dma_callback;
audio_rate_div=audio_rate/1000;

snd_set0w((s32 *) audio_buf[0], SND_BUFFERSIZE>>1);
snd_set0w((s32 *) audio_buf[1], SND_BUFFERSIZE>>1);

if ( SDL_OpenAudio(&audio, NULL) < 0 ) 
   {
   printf("SDL OpenAudio failed\n");
   exit(1);
   } 
 audio_rate= audio.freq; 
SDL_PauseAudio(1);
#else
if(samplerate==0) {audio_rate=32000;AUDIO_SetDSPSampleRate(AI_SAMPLERATE_32KHZ);}
else {audio_rate=48000;AUDIO_SetDSPSampleRate(AI_SAMPLERATE_48KHZ);}


audio_rate_div=audio_rate/1000;

snd_set0w((s32 *) audio_buf[0], SND_BUFFERSIZE>>1);
snd_set0w((s32 *) audio_buf[1], SND_BUFFERSIZE>>1);

DCFlushRange(audio_buf[0],SND_BUFFERSIZE<<1);
DCFlushRange(audio_buf[1],SND_BUFFERSIZE<<1);

AUDIO_RegisterDMACallback(audio_dma_callback);
#endif

}


void SND_End()
{
if(audio_init==0) return;
audio_init=0;

#ifdef _SDL_LENDIAN_ 
    if(audio_paused==0) SDL_PauseAudio(1);   
      audio_paused=1;
#else
     if(audio_paused==0) AUDIO_StopDMA();
     AUDIO_RegisterDMACallback(NULL);
     audio_paused=1;
#endif

}

s32 SND_Is_Paused()
{
return audio_paused;
}

void SND_Pause(s32 paused)
{
if(audio_init==0) return;

if(paused)
	{
    #ifdef _SDL_LENDIAN_ 
    if(audio_paused==0) SDL_PauseAudio(1);    
    #else
    if(audio_paused==0) AUDIO_StopDMA();
    #endif
	audio_paused=1;
	}
else
	{
	if(audio_paused!=0)
	{
	curr_audio_buf=0;
	snd_set0w((s32 *) audio_buf[curr_audio_buf], SND_BUFFERSIZE>>1);
	#ifdef _SDL_LENDIAN_	
    SDL_PauseAudio(0);
    #else
    AUDIO_InitDMA((u32)audio_buf[curr_audio_buf],SND_BUFFERSIZE<<1);
	AUDIO_StartDMA();
	#endif
    curr_audio_buf ^= 1;
	mix_voices(audio_buf[curr_audio_buf], SND_BUFFERSIZE>>1);
    #ifdef _SDL_LENDIAN_ 
    
    #else
    DCFlushRange(audio_buf[curr_audio_buf],SND_BUFFERSIZE<<1);
    #endif
    
	audio_paused=0;
	}
	
	}
}

u32 SND_GetSampleCounter()
{
	
return sample_counter;
}

u32 SND_GetSamplesPerTick()
{
return (SND_BUFFERSIZE>>1);
}
	
u32 SND_GetTime()
{
	
return (sample_counter/ (u32) audio_rate_div);
}

void SND_SetTime(u32 time)
{
	
 sample_counter= ((u32) audio_rate_div)*time;

}

void SND_SetCallback(void (*callback)())
{

 snd_callback=callback;

}


// VOICE FUNCTIONS


s32 SND_SetVoice(s32 voice, s32 format, s32 pitch,s32 delay, void *snd, s32 size_snd, s32 volume_l, s32 volume_r, void (*callback) (s32 voice))
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice

    if(size_snd<=0 || snd==NULL) return -1; // invalid voice

	if(delay<0) delay=0;

	delay= audio_rate_div * delay;

	if(pitch<1) pitch=1;
	if(pitch>MAX_PITCH) pitch=MAX_PITCH;

	volume_l &=255;
	volume_r &=255;

	voices[voice].flags=format & 3;
	
	if(callback) voices[voice].flags|=VOICE_DOUBLEBUFFER;

	voices[voice].voice= snd;
    voices[voice].voice2= snd;
	voices[voice].size= size_snd;
	voices[voice].size2= 0;
	voices[voice].delay= delay;
	voices[voice].volume_l= volume_l;
	voices[voice].volume_r= volume_r; 
	voices[voice].callback= callback;

	voices[voice].pitch= pitch;
	voices[voice].pos= 0;
	voices[voice].sub_pos= 0;
	
	voices[voice].tick_counter=0;

	voices[voice].old_sample_l=voices[voice].old_sample_r=0;
	
	voices[voice].flags|= VOICE_ACTIVE;
   

return 0;
}

s32 SND_AddVoice(s32 voice, void *snd, s32 size_snd)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	
	if(!(voices[voice].flags & VOICE_ACTIVE)) return -1; // invalid voice
	
	if(size_snd<=0 || snd==NULL) return -1; // invalid voice
	
	if(voices[voice].size2>0) return 1; // busy
    voices[voice].voice2= snd;
	voices[voice].size2=size_snd;


return 0;
}

s32 SND_StopVoice(s32 voice)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	voices[voice].flags&= ~VOICE_ACTIVE;

return 0;
}

s32 SND_PauseVoice(s32 voice, s32 pause)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	if(pause) voices[voice].flags|=VOICE_SONGHALT; else voices[voice].flags&= ~VOICE_SONGHALT;

return 0;
}

s32 SND_StatusVoice(s32 voice)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	
	if(voices[voice].flags & VOICE_ACTIVE) 
	 {
	 if(!(voices[voice].flags & VOICE_SONGSAMPLE))
		{
	     if(voices[voice].size<=0 && voices[voice].size2<=0) return 2; // voice paused
		}
	 else 
		{
		 if(!(voices[voice].flags & VOICE_SONGHALT))  return 2; // voice paused
		}
	
	  return 1; // voice playing
	 }

return 0; // voice disabled
}

s32 SND_GetFirstUnusedVoice()
{
s32 n;	
	
	for(n=1;n<MAX_SND_VOICES;n++)
	   if(!(voices[n].flags & VOICE_ACTIVE)) return n;
    
	if(!(voices[0].flags & VOICE_ACTIVE)) return 0; // voice 0 is a special case

return -1;  // all voices used
}

// special

s32 SND_ChangePitchVoice(s32 voice, s32 pitch)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	
	if(!(voices[voice].flags & VOICE_ACTIVE)) return -1;  // invalid voice

	if(pitch<1) pitch=1;
	if(pitch>144000) pitch=144000;

	voices[voice].pitch= pitch;


return 0;
}

s32 SND_ChangeVolumeVoice(s32 voice, s32 volume_l, s32 volume_r)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	
	if(!(voices[voice].flags & VOICE_ACTIVE)) return -1;  // invalid voice

	volume_l &=255;
	volume_r &=255;

	voices[voice].volume_l= volume_l;
	voices[voice].volume_r= volume_r;

return 0;
}

u32 SND_GetTickCounterVoice(s32 voice)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return 0; 
	
return voices[voice].tick_counter;
}

u32 SND_GetTimerVoice(s32 voice)
{
	if(voice<0 || voice>=MAX_SND_VOICES) return 0; // invalid voice
	
return (voices[voice].tick_counter / ((u32) audio_rate_div));
}

s32 SND_TestPointer(s32 voice, void *pointer)
{
if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice

if(voices[voice].size>0 && voices[voice].voice==pointer) return 1;
if(voices[voice].size2>0 && voices[voice].voice2==pointer) return 1;

return 0;
}


// Music funtions

s32 SND_PlaySongVoice(s32 voice, s32 pitch,s32 delay, s32 attack, s32 sustain, s32 decay, s32 volume_l, s32 volume_r)
{
    if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	if(!(voices[voice].flags & VOICE_ACTIVE)) return -1;  // invalid voice
	if(!(voices[voice].flags & VOICE_SONGSAMPLE)) return -2;  // invalid voice for song

    if(delay<0) delay=0;

	

	if(pitch<1) pitch=1;
	if(pitch>MAX_PITCH) pitch=MAX_PITCH;

	delay= audio_rate_div * delay;

    // milliseconds to samples

	if(sustain>0)
	    sustain= (s32) (((s64) pitch)*((s64) sustain)/1000LL);
	else sustain=0;
   

    if(attack>0)
		attack=(s32)(0x3E7FFFC18LL/(((s64) pitch)*((s64) attack)));
	else attack=0;
	   
	if(decay>0)
		decay= (s32)(0x3E7FFFC18LL/(((s64) pitch)*((s64) decay )));
	else decay=0;

	volume_l &=255;
	volume_r &=255;

    if(pitch>audio_rate)
	 {
	 if(sustain)
		sustain= (s32) ((((s64) sustain)*((s64) audio_rate))/((s64) pitch));
     
	 if(attack)
		attack = (s32) ((((s64) attack )*((s64) pitch))/((s64) audio_rate));
	 
	 if(decay)
		decay  = (s32) ((((s64) decay  )*((s64) pitch))/((s64) audio_rate));
	 }

    voices[voice].flags|=VOICE_SONGHALT;
	
	voices[voice].delay= delay;

	voices[voice].volume_l= volume_l;
	voices[voice].volume_r= volume_r;

	voices[voice].voice= voices[voice].voice_aux;
	voices[voice].size=  voices[voice].size_aux;

	voices[voice].pitch= pitch;
	voices[voice].pos= 0;
	voices[voice].sub_pos= 0;

	voices[voice].ticks_for_attack=0xffffff*(attack>0);
    voices[voice].ticks_sub_for_attack=attack;

    voices[voice].ticks_for_sustain=sustain;

    voices[voice].ticks_for_decay=0xffffff*(decay>0);
    voices[voice].ticks_sub_for_decay= decay;

	voices[voice].tick_counter=0;

	voices[voice].old_sample_l=voices[voice].old_sample_r=0;

    voices[voice].flags&=~VOICE_SONGHALT;

return 0;
}

s32 SND_SetSongSampleVoice(s32 voice, s32 format, void *snd_start, void *snd_start_loop, s32 size_snd)
{

	if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice

    if(size_snd<=0 || snd_start==NULL || (format & 2)==2) return -1; // invalid voice

	
    if(snd_start_loop==NULL || snd_start_loop<snd_start) {snd_start_loop=NULL;}

	voices[voice].flags=format & 3;
	
	voices[voice].flags|=VOICE_DOUBLEBUFFER | VOICE_SONGSAMPLE | VOICE_SONGHALT;

	voices[voice].voice= snd_start;
    voices[voice].voice2= snd_start_loop;
	voices[voice].size= size_snd;
	if(snd_start_loop)
      voices[voice].size2= size_snd-(((int) snd_start_loop)-((int) snd_start));
    else  voices[voice].size2=0;
    
	voices[voice].voice_aux= snd_start;
	voices[voice].size_aux= size_snd;

	voices[voice].delay= 0;

	voices[voice].volume_l= 255;
	voices[voice].volume_r= 255;

	voices[voice].callback= NULL;

	voices[voice].pitch= audio_rate;
	voices[voice].pos= 0;
	voices[voice].sub_pos= 0;

	voices[voice].ticks_for_attack=0;
    voices[voice].ticks_sub_for_attack=0;

    voices[voice].ticks_for_sustain=0;

    voices[voice].ticks_for_decay=0;
    voices[voice].ticks_sub_for_decay=0;
	
	voices[voice].tick_counter=0;

	voices[voice].flags|= VOICE_ACTIVE;
    
return 0;
}

s32 SND_PauseSongVoice(s32 voice, s32 pause)
{
    if(voice<0 || voice>=MAX_SND_VOICES) return -1; // invalid voice
	if(!(voices[voice].flags & VOICE_ACTIVE)) return -1;  // invalid voice
	if(!(voices[voice].flags & VOICE_SONGSAMPLE)) return -2;  // invalid voice for song

    
    if(pause) voices[voice].flags|=VOICE_SONGHALT;
	else voices[voice].flags&=~VOICE_SONGHALT;
	
return 0;
}


int Note2Freq(int note, int freq_base,int note_base)
{
int n;
static int one=1;
static u32 tab_piano_frac[12];

if(one)
{
float note=1.0f;
one=0;
for(n=0;n<12;n++) // table note
{
tab_piano_frac[n]=(u32)(10000.0f*note);
note*=1.0594386f;
}
}

// obtiene octava 3 (notas 36 a 47)

n=(note/12)-(note_base/12);
if(n>=0) freq_base<<=n;
else freq_base>>= -n;

//if(freq_base<=0x1fffff) // Math precision
if(freq_base<=0x1ffff) // Math precision
  n=(s32) (((u32)freq_base)*tab_piano_frac[(note % 12)]/tab_piano_frac[(note_base % 12)]);
else 
  n=(s32) (((u64)freq_base)*((u64) tab_piano_frac[(note % 12)])/((u64) tab_piano_frac[(note_base % 12)]));
  

return n;
}


/************************** END **************************/

